# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/yuya/work/hghacks/thg/tortoisehg/hgqt/postreview.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from tortoisehg.util.i18n import _
from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_PostReviewDialog(object):
    def setupUi(self, PostReviewDialog):
        PostReviewDialog.setObjectName("PostReviewDialog")
        PostReviewDialog.resize(660, 459)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(PostReviewDialog)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.tab_widget = QtWidgets.QTabWidget(PostReviewDialog)
        self.tab_widget.setMaximumSize(QtCore.QSize(16777215, 110))
        self.tab_widget.setObjectName("tab_widget")
        self.post_review_tab = QtWidgets.QWidget()
        self.post_review_tab.setObjectName("post_review_tab")
        self.formLayout_2 = QtWidgets.QFormLayout(self.post_review_tab)
        self.formLayout_2.setObjectName("formLayout_2")
        self.repo_id_label = QtWidgets.QLabel(self.post_review_tab)
        self.repo_id_label.setObjectName("repo_id_label")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.repo_id_label)
        self.repo_id_combo = QtWidgets.QComboBox(self.post_review_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.repo_id_combo.sizePolicy().hasHeightForWidth())
        self.repo_id_combo.setSizePolicy(sizePolicy)
        self.repo_id_combo.setEditable(False)
        self.repo_id_combo.setInsertPolicy(QtWidgets.QComboBox.InsertAtTop)
        self.repo_id_combo.setObjectName("repo_id_combo")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.repo_id_combo)
        self.summary_label = QtWidgets.QLabel(self.post_review_tab)
        self.summary_label.setObjectName("summary_label")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.summary_label)
        self.summary_edit = QtWidgets.QComboBox(self.post_review_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.summary_edit.sizePolicy().hasHeightForWidth())
        self.summary_edit.setSizePolicy(sizePolicy)
        self.summary_edit.setEditable(True)
        self.summary_edit.setInsertPolicy(QtWidgets.QComboBox.InsertAtTop)
        self.summary_edit.setObjectName("summary_edit")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.summary_edit)
        self.tab_widget.addTab(self.post_review_tab, "")
        self.update_review_tab = QtWidgets.QWidget()
        self.update_review_tab.setObjectName("update_review_tab")
        self.formLayout_3 = QtWidgets.QFormLayout(self.update_review_tab)
        self.formLayout_3.setObjectName("formLayout_3")
        self.review_id_label = QtWidgets.QLabel(self.update_review_tab)
        self.review_id_label.setObjectName("review_id_label")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.review_id_label)
        self.review_id_combo = QtWidgets.QComboBox(self.update_review_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.review_id_combo.sizePolicy().hasHeightForWidth())
        self.review_id_combo.setSizePolicy(sizePolicy)
        self.review_id_combo.setEditable(False)
        self.review_id_combo.setInsertPolicy(QtWidgets.QComboBox.InsertAtTop)
        self.review_id_combo.setObjectName("review_id_combo")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.review_id_combo)
        self.update_fields = QtWidgets.QCheckBox(self.update_review_tab)
        self.update_fields.setObjectName("update_fields")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.update_fields)
        self.tab_widget.addTab(self.update_review_tab, "")
        self.verticalLayout.addWidget(self.tab_widget)
        self.options_group = QtWidgets.QGroupBox(PostReviewDialog)
        self.options_group.setObjectName("options_group")
        self.gridLayout = QtWidgets.QGridLayout(self.options_group)
        self.gridLayout.setObjectName("gridLayout")
        self.outgoing_changes_check = QtWidgets.QCheckBox(self.options_group)
        self.outgoing_changes_check.setObjectName("outgoing_changes_check")
        self.gridLayout.addWidget(self.outgoing_changes_check, 0, 0, 1, 1)
        self.branch_check = QtWidgets.QCheckBox(self.options_group)
        self.branch_check.setObjectName("branch_check")
        self.gridLayout.addWidget(self.branch_check, 0, 1, 1, 1)
        self.publish_immediately_check = QtWidgets.QCheckBox(self.options_group)
        self.publish_immediately_check.setObjectName("publish_immediately_check")
        self.gridLayout.addWidget(self.publish_immediately_check, 2, 0, 1, 1)
        self.verticalLayout.addWidget(self.options_group)
        self.changesets_box = QtWidgets.QGroupBox(PostReviewDialog)
        self.changesets_box.setEnabled(True)
        self.changesets_box.setToolTip("")
        self.changesets_box.setObjectName("changesets_box")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.changesets_box)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.changesets_view = QtWidgets.QTreeView(self.changesets_box)
        self.changesets_view.setIndentation(0)
        self.changesets_view.setRootIsDecorated(False)
        self.changesets_view.setItemsExpandable(False)
        self.changesets_view.setObjectName("changesets_view")
        self.changesets_view.header().setHighlightSections(False)
        self.changesets_view.header().setSortIndicatorShown(False)
        self.verticalLayout_3.addWidget(self.changesets_view)
        self.verticalLayout.addWidget(self.changesets_box)
        self.verticalLayout_5.addLayout(self.verticalLayout)
        self.dialogbuttons_layout = QtWidgets.QHBoxLayout()
        self.dialogbuttons_layout.setObjectName("dialogbuttons_layout")
        self.settings_button = QtWidgets.QPushButton(PostReviewDialog)
        self.settings_button.setToolTip("")
        self.settings_button.setDefault(False)
        self.settings_button.setObjectName("settings_button")
        self.dialogbuttons_layout.addWidget(self.settings_button)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.dialogbuttons_layout.addItem(spacerItem)
        self.progress_bar = QtWidgets.QProgressBar(PostReviewDialog)
        self.progress_bar.setMinimumSize(QtCore.QSize(200, 0))
        font = QtGui.QFont()
        font.setKerning(True)
        self.progress_bar.setFont(font)
        self.progress_bar.setMinimum(0)
        self.progress_bar.setMaximum(0)
        self.progress_bar.setProperty("value", -1)
        self.progress_bar.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.progress_bar.setTextVisible(True)
        self.progress_bar.setOrientation(QtCore.Qt.Horizontal)
        self.progress_bar.setInvertedAppearance(False)
        self.progress_bar.setTextDirection(QtWidgets.QProgressBar.TopToBottom)
        self.progress_bar.setObjectName("progress_bar")
        self.dialogbuttons_layout.addWidget(self.progress_bar)
        self.progress_label = QtWidgets.QLabel(PostReviewDialog)
        self.progress_label.setObjectName("progress_label")
        self.dialogbuttons_layout.addWidget(self.progress_label)
        spacerItem1 = QtWidgets.QSpacerItem(0, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.dialogbuttons_layout.addItem(spacerItem1)
        self.post_review_button = QtWidgets.QPushButton(PostReviewDialog)
        self.post_review_button.setEnabled(False)
        self.post_review_button.setDefault(False)
        self.post_review_button.setObjectName("post_review_button")
        self.dialogbuttons_layout.addWidget(self.post_review_button)
        self.close_button = QtWidgets.QPushButton(PostReviewDialog)
        self.close_button.setEnabled(True)
        self.close_button.setDefault(True)
        self.close_button.setObjectName("close_button")
        self.dialogbuttons_layout.addWidget(self.close_button)
        self.verticalLayout_5.addLayout(self.dialogbuttons_layout)
        self.repo_id_label.setBuddy(self.repo_id_combo)
        self.summary_label.setBuddy(self.summary_edit)
        self.review_id_label.setBuddy(self.review_id_combo)

        self.retranslateUi(PostReviewDialog)
        self.tab_widget.setCurrentIndex(0)
        self.post_review_button.clicked.connect(PostReviewDialog.accept)
        self.settings_button.clicked.connect(PostReviewDialog.onSettingsButtonClicked)
        self.close_button.clicked.connect(PostReviewDialog.close)
        self.outgoing_changes_check.toggled['bool'].connect(PostReviewDialog.outgoingChangesCheckToggle)
        self.branch_check.toggled['bool'].connect(PostReviewDialog.branchCheckToggle)
        self.tab_widget.currentChanged['int'].connect(PostReviewDialog.tabChanged)
        QtCore.QMetaObject.connectSlotsByName(PostReviewDialog)
        PostReviewDialog.setTabOrder(self.changesets_view, self.post_review_button)
        PostReviewDialog.setTabOrder(self.post_review_button, self.settings_button)

    def retranslateUi(self, PostReviewDialog):
        _translate = QtCore.QCoreApplication.translate
        PostReviewDialog.setWindowTitle(_("Review Board"))
        self.repo_id_label.setText(_("Repository ID:"))
        self.summary_label.setText(_("Summary:"))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.post_review_tab), _("Post Review"))
        self.review_id_label.setText(_("Review ID:"))
        self.update_fields.setText(_("Update the fields of this existing request"))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.update_review_tab), _("Update Review"))
        self.options_group.setTitle(_("Options"))
        self.outgoing_changes_check.setText(_("Create diff with all outgoing changes"))
        self.branch_check.setText(_("Create diff with all changes on this branch"))
        self.publish_immediately_check.setText(_("Publish request immediately"))
        self.changesets_box.setTitle(_("Changesets"))
        self.settings_button.setText(_("&Settings"))
        self.progress_bar.setFormat(_("%p%"))
        self.progress_label.setText(_("Connecting to Review Board..."))
        self.post_review_button.setText(_("Post &Review"))
        self.close_button.setText(_("&Close"))
