# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/yuya/work/hghacks/thg/tortoisehg/hgqt/hgemail.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from tortoisehg.util.i18n import _
from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_EmailDialog(object):
    def setupUi(self, EmailDialog):
        EmailDialog.setObjectName("EmailDialog")
        EmailDialog.resize(660, 519)
        EmailDialog.setSizeGripEnabled(True)
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(EmailDialog)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.main_tabs = QtWidgets.QTabWidget(EmailDialog)
        self.main_tabs.setDocumentMode(False)
        self.main_tabs.setTabsClosable(False)
        self.main_tabs.setMovable(False)
        self.main_tabs.setObjectName("main_tabs")
        self.edit_tab = QtWidgets.QWidget()
        self.edit_tab.setObjectName("edit_tab")
        self.gridLayout = QtWidgets.QGridLayout(self.edit_tab)
        self.gridLayout.setObjectName("gridLayout")
        self.envelope_box = QtWidgets.QGroupBox(self.edit_tab)
        self.envelope_box.setTitle("")
        self.envelope_box.setObjectName("envelope_box")
        self.formLayout = QtWidgets.QFormLayout(self.envelope_box)
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.to_label = QtWidgets.QLabel(self.envelope_box)
        self.to_label.setObjectName("to_label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.to_label)
        self.to_edit = QtWidgets.QComboBox(self.envelope_box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.to_edit.sizePolicy().hasHeightForWidth())
        self.to_edit.setSizePolicy(sizePolicy)
        self.to_edit.setEditable(True)
        self.to_edit.setInsertPolicy(QtWidgets.QComboBox.InsertAtTop)
        self.to_edit.setObjectName("to_edit")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.to_edit)
        self.cc_label = QtWidgets.QLabel(self.envelope_box)
        self.cc_label.setObjectName("cc_label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.cc_label)
        self.cc_edit = QtWidgets.QComboBox(self.envelope_box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cc_edit.sizePolicy().hasHeightForWidth())
        self.cc_edit.setSizePolicy(sizePolicy)
        self.cc_edit.setEditable(True)
        self.cc_edit.setInsertPolicy(QtWidgets.QComboBox.InsertAtTop)
        self.cc_edit.setObjectName("cc_edit")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.cc_edit)
        self.from_label = QtWidgets.QLabel(self.envelope_box)
        self.from_label.setObjectName("from_label")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.from_label)
        self.from_edit = QtWidgets.QComboBox(self.envelope_box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.from_edit.sizePolicy().hasHeightForWidth())
        self.from_edit.setSizePolicy(sizePolicy)
        self.from_edit.setEditable(True)
        self.from_edit.setInsertPolicy(QtWidgets.QComboBox.InsertAtTop)
        self.from_edit.setObjectName("from_edit")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.from_edit)
        self.inreplyto_label = QtWidgets.QLabel(self.envelope_box)
        self.inreplyto_label.setObjectName("inreplyto_label")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.inreplyto_label)
        self.inreplyto_edit = QtWidgets.QLineEdit(self.envelope_box)
        self.inreplyto_edit.setObjectName("inreplyto_edit")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.inreplyto_edit)
        self.flag_label = QtWidgets.QLabel(self.envelope_box)
        self.flag_label.setObjectName("flag_label")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.flag_label)
        self.flag_edit = QtWidgets.QComboBox(self.envelope_box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.flag_edit.sizePolicy().hasHeightForWidth())
        self.flag_edit.setSizePolicy(sizePolicy)
        self.flag_edit.setEditable(True)
        self.flag_edit.setInsertPolicy(QtWidgets.QComboBox.InsertAtTop)
        self.flag_edit.setObjectName("flag_edit")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.flag_edit)
        self.gridLayout.addWidget(self.envelope_box, 0, 0, 1, 1)
        self.options_edit = QtWidgets.QGroupBox(self.edit_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.options_edit.sizePolicy().hasHeightForWidth())
        self.options_edit.setSizePolicy(sizePolicy)
        self.options_edit.setTitle("")
        self.options_edit.setObjectName("options_edit")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.options_edit)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.patch_frame = QtWidgets.QFrame(self.options_edit)
        self.patch_frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.patch_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.patch_frame.setObjectName("patch_frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.patch_frame)
        self.verticalLayout.setObjectName("verticalLayout")
        self.hgpatch_radio = QtWidgets.QRadioButton(self.patch_frame)
        self.hgpatch_radio.setObjectName("hgpatch_radio")
        self.verticalLayout.addWidget(self.hgpatch_radio)
        self.gitpatch_radio = QtWidgets.QRadioButton(self.patch_frame)
        self.gitpatch_radio.setObjectName("gitpatch_radio")
        self.verticalLayout.addWidget(self.gitpatch_radio)
        self.plainpatch_radio = QtWidgets.QRadioButton(self.patch_frame)
        self.plainpatch_radio.setObjectName("plainpatch_radio")
        self.verticalLayout.addWidget(self.plainpatch_radio)
        self.bundle_radio = QtWidgets.QRadioButton(self.patch_frame)
        self.bundle_radio.setObjectName("bundle_radio")
        self.verticalLayout.addWidget(self.bundle_radio)
        self.verticalLayout_4.addWidget(self.patch_frame)
        self.extra_frame = QtWidgets.QFrame(self.options_edit)
        self.extra_frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.extra_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.extra_frame.setObjectName("extra_frame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.extra_frame)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.body_check = QtWidgets.QCheckBox(self.extra_frame)
        self.body_check.setEnabled(True)
        self.body_check.setChecked(True)
        self.body_check.setObjectName("body_check")
        self.horizontalLayout.addWidget(self.body_check)
        self.attach_check = QtWidgets.QCheckBox(self.extra_frame)
        self.attach_check.setObjectName("attach_check")
        self.horizontalLayout.addWidget(self.attach_check)
        self.inline_check = QtWidgets.QCheckBox(self.extra_frame)
        self.inline_check.setObjectName("inline_check")
        self.horizontalLayout.addWidget(self.inline_check)
        self.diffstat_check = QtWidgets.QCheckBox(self.extra_frame)
        self.diffstat_check.setObjectName("diffstat_check")
        self.horizontalLayout.addWidget(self.diffstat_check)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout_4.addWidget(self.extra_frame)
        self.gridLayout.addWidget(self.options_edit, 0, 1, 1, 1)
        self.writeintro_check = QtWidgets.QCheckBox(self.edit_tab)
        self.writeintro_check.setObjectName("writeintro_check")
        self.gridLayout.addWidget(self.writeintro_check, 1, 0, 1, 2)
        self.intro_changesets_splitter = QtWidgets.QSplitter(self.edit_tab)
        self.intro_changesets_splitter.setOrientation(QtCore.Qt.Vertical)
        self.intro_changesets_splitter.setObjectName("intro_changesets_splitter")
        self.intro_box = QtWidgets.QGroupBox(self.intro_changesets_splitter)
        self.intro_box.setTitle("")
        self.intro_box.setObjectName("intro_box")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.intro_box)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.subject_layout = QtWidgets.QHBoxLayout()
        self.subject_layout.setObjectName("subject_layout")
        self.subject_label = QtWidgets.QLabel(self.intro_box)
        self.subject_label.setObjectName("subject_label")
        self.subject_layout.addWidget(self.subject_label)
        self.subject_edit = QtWidgets.QComboBox(self.intro_box)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.subject_edit.sizePolicy().hasHeightForWidth())
        self.subject_edit.setSizePolicy(sizePolicy)
        self.subject_edit.setEditable(True)
        self.subject_edit.setInsertPolicy(QtWidgets.QComboBox.InsertAtTop)
        self.subject_edit.setObjectName("subject_edit")
        self.subject_layout.addWidget(self.subject_edit)
        self.verticalLayout_2.addLayout(self.subject_layout)
        self.body_edit = QtWidgets.QPlainTextEdit(self.intro_box)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.body_edit.setFont(font)
        self.body_edit.setObjectName("body_edit")
        self.verticalLayout_2.addWidget(self.body_edit)
        self.changesets_box = QtWidgets.QGroupBox(self.intro_changesets_splitter)
        self.changesets_box.setObjectName("changesets_box")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.changesets_box)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.changesets_view = QtWidgets.QTreeView(self.changesets_box)
        self.changesets_view.setIndentation(0)
        self.changesets_view.setRootIsDecorated(False)
        self.changesets_view.setItemsExpandable(False)
        self.changesets_view.setObjectName("changesets_view")
        self.verticalLayout_3.addWidget(self.changesets_view)
        self.selectallnone_layout = QtWidgets.QHBoxLayout()
        self.selectallnone_layout.setObjectName("selectallnone_layout")
        self.selectall_button = QtWidgets.QPushButton(self.changesets_box)
        self.selectall_button.setObjectName("selectall_button")
        self.selectallnone_layout.addWidget(self.selectall_button)
        self.selectnone_button = QtWidgets.QPushButton(self.changesets_box)
        self.selectnone_button.setObjectName("selectnone_button")
        self.selectallnone_layout.addWidget(self.selectnone_button)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.selectallnone_layout.addItem(spacerItem1)
        self.verticalLayout_3.addLayout(self.selectallnone_layout)
        self.gridLayout.addWidget(self.intro_changesets_splitter, 2, 0, 1, 2)
        self.main_tabs.addTab(self.edit_tab, "")
        self.preview_tab = QtWidgets.QWidget()
        self.preview_tab.setObjectName("preview_tab")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.preview_tab)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.preview_edit = Qsci.QsciScintilla(self.preview_tab)
        self.preview_edit.setObjectName("preview_edit")
        self.gridLayout_2.addWidget(self.preview_edit, 0, 0, 1, 1)
        self.main_tabs.addTab(self.preview_tab, "")
        self.verticalLayout_5.addWidget(self.main_tabs)
        self.dialogbuttons_layout = QtWidgets.QHBoxLayout()
        self.dialogbuttons_layout.setObjectName("dialogbuttons_layout")
        self.settings_button = QtWidgets.QPushButton(EmailDialog)
        self.settings_button.setToolTip("")
        self.settings_button.setDefault(False)
        self.settings_button.setObjectName("settings_button")
        self.dialogbuttons_layout.addWidget(self.settings_button)
        spacerItem2 = QtWidgets.QSpacerItem(25, 19, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.dialogbuttons_layout.addItem(spacerItem2)
        self.send_button = QtWidgets.QPushButton(EmailDialog)
        self.send_button.setEnabled(False)
        self.send_button.setDefault(False)
        self.send_button.setObjectName("send_button")
        self.dialogbuttons_layout.addWidget(self.send_button)
        self.close_button = QtWidgets.QPushButton(EmailDialog)
        self.close_button.setEnabled(True)
        self.close_button.setDefault(True)
        self.close_button.setObjectName("close_button")
        self.dialogbuttons_layout.addWidget(self.close_button)
        self.verticalLayout_5.addLayout(self.dialogbuttons_layout)
        self.to_label.setBuddy(self.to_edit)
        self.cc_label.setBuddy(self.cc_edit)
        self.from_label.setBuddy(self.from_edit)
        self.inreplyto_label.setBuddy(self.inreplyto_edit)
        self.flag_label.setBuddy(self.flag_edit)
        self.subject_label.setBuddy(self.subject_edit)

        self.retranslateUi(EmailDialog)
        self.main_tabs.setCurrentIndex(0)
        self.writeintro_check.toggled['bool'].connect(self.intro_box.setVisible)
        self.send_button.clicked.connect(EmailDialog.accept)
        self.close_button.clicked.connect(EmailDialog.close)
        self.writeintro_check.toggled['bool'].connect(self.subject_edit.setFocus)
        QtCore.QMetaObject.connectSlotsByName(EmailDialog)
        EmailDialog.setTabOrder(self.main_tabs, self.to_edit)
        EmailDialog.setTabOrder(self.to_edit, self.cc_edit)
        EmailDialog.setTabOrder(self.cc_edit, self.from_edit)
        EmailDialog.setTabOrder(self.from_edit, self.inreplyto_edit)
        EmailDialog.setTabOrder(self.inreplyto_edit, self.flag_edit)
        EmailDialog.setTabOrder(self.flag_edit, self.hgpatch_radio)
        EmailDialog.setTabOrder(self.hgpatch_radio, self.gitpatch_radio)
        EmailDialog.setTabOrder(self.gitpatch_radio, self.plainpatch_radio)
        EmailDialog.setTabOrder(self.plainpatch_radio, self.bundle_radio)
        EmailDialog.setTabOrder(self.bundle_radio, self.body_check)
        EmailDialog.setTabOrder(self.body_check, self.attach_check)
        EmailDialog.setTabOrder(self.attach_check, self.inline_check)
        EmailDialog.setTabOrder(self.inline_check, self.diffstat_check)
        EmailDialog.setTabOrder(self.diffstat_check, self.writeintro_check)
        EmailDialog.setTabOrder(self.writeintro_check, self.subject_edit)
        EmailDialog.setTabOrder(self.subject_edit, self.body_edit)
        EmailDialog.setTabOrder(self.body_edit, self.changesets_view)
        EmailDialog.setTabOrder(self.changesets_view, self.send_button)
        EmailDialog.setTabOrder(self.send_button, self.preview_edit)
        EmailDialog.setTabOrder(self.preview_edit, self.settings_button)

    def retranslateUi(self, EmailDialog):
        _translate = QtCore.QCoreApplication.translate
        EmailDialog.setWindowTitle(_("Email"))
        self.to_label.setText(_("To:"))
        self.cc_label.setText(_("Cc:"))
        self.from_label.setText(_("From:"))
        self.inreplyto_label.setText(_("In-Reply-To:"))
        self.inreplyto_edit.setToolTip(_("Message identifier to reply to, for threading"))
        self.flag_label.setText(_("Flag:"))
        self.hgpatch_radio.setWhatsThis(_("Hg patches (as generated by export command) are compatible with most patch programs.  They include a header which contains the most important changeset metadata."))
        self.hgpatch_radio.setText(_("Send changesets as Hg patches"))
        self.gitpatch_radio.setWhatsThis(_("Git patches can describe binary files, copies, and permission changes, but recipients may not be able to use them if they are not using git or Mercurial."))
        self.gitpatch_radio.setText(_("Use extended (git) patch format"))
        self.plainpatch_radio.setWhatsThis(_("Stripping Mercurial header removes username and parent information.  Only useful if recipient is not using Mercurial (and does not like to see the headers)."))
        self.plainpatch_radio.setText(_("Plain, do not prepend Hg header"))
        self.bundle_radio.setWhatsThis(_("Bundles store complete changesets in binary form. Upstream users can pull from them. This is the safest way to send changes to recipient Mercurial users."))
        self.bundle_radio.setText(_("Send single binary bundle, not patches"))
        self.body_check.setToolTip(_("send patches as part of the email body"))
        self.body_check.setText(_("body"))
        self.attach_check.setToolTip(_("send patches as attachments"))
        self.attach_check.setText(_("attach"))
        self.inline_check.setToolTip(_("send patches as inline attachments"))
        self.inline_check.setText(_("inline"))
        self.diffstat_check.setToolTip(_("add diffstat output to messages"))
        self.diffstat_check.setText(_("diffstat"))
        self.writeintro_check.setWhatsThis(_("Patch series description is sent in initial summary email with [PATCH 0 of N] subject.  It should describe the effects of the entire patch series.  When emailing a bundle, these fields make up the message subject and body. Flags is a comma separated list of tags which are inserted into the message subject prefix."))
        self.writeintro_check.setText(_("Write patch series (bundle) description"))
        self.subject_label.setText(_("Subject:"))
        self.changesets_box.setTitle(_("Changesets"))
        self.selectall_button.setText(_("Select &All"))
        self.selectnone_button.setText(_("Select &None"))
        self.main_tabs.setTabText(self.main_tabs.indexOf(self.edit_tab), _("Edit"))
        self.main_tabs.setTabText(self.main_tabs.indexOf(self.preview_tab), _("Preview"))
        self.settings_button.setText(_("&Settings"))
        self.send_button.setText(_("Send &Email"))
        self.close_button.setText(_("&Close"))
from PyQt5 import Qsci
